"注入式执行函数"

import time
from typing import TYPE_CHECKING, Any, Optional
from ...color_print import Print
from ...packets import Packet_CommandOutput
from ...game_utils import getTarget, getPos

if TYPE_CHECKING:
    from tooldelta.frame import ToolDelta, GameCtrl


def check_avaliable(sth: "GameCtrl") -> Optional[AttributeError]:
    """
    检查给定的 "GameCtrl" 对象是否可用

    参数:
        sth: 要检查的 "GameCtrl" 对象
    """
    if sth is None:
        raise AttributeError(f"无法使用 {sth.__class__.__name__}, 因为其还未被初始化")


game_control: "GameCtrl"
movent_frame: "ToolDelta"


def set_frame(my_Frame: "ToolDelta") -> None:
    """
    全局初始化框架

    参数:
        my_Frame: 要设置的框架对象
    """
    global movent_frame, game_control  # pylint: disable=global-statement
    movent_frame = my_Frame
    game_control = my_Frame.get_game_control()


def sendcmd(
    cmd: str, waitForResp: bool = False, timeout: int = 30
) -> None | Packet_CommandOutput:
    r"""发送命令到游戏控制器，并可选择是否等待响应

    如果 waitForResp 为 False，则返回 None，否则返回 Packet_CommandOutput 对象

    参数:
        cmd: 要发送的命令
        waitForResp: 是否等待响应，默认为 False
        timeout: 等待响应的超时时间（秒）,默认为 30
    """
    check_avaliable(game_control)
    return game_control.sendcmd(cmd, waitForResp, timeout)


def sendwscmd(
    cmd: str, waitForResp: bool = False, timeout: float = 30
) -> Packet_CommandOutput | None:
    """
    发送 WSCMD 命令到游戏控制器

    参数:
        cmd: 要发送的 WSCMD 命令
        waitForResp: 是否等待响应 默认为 False
        timeout: 超时时间（秒）默认为 30
    """
    check_avaliable(game_control)
    return game_control.sendwscmd(cmd, waitForResp, timeout)


def sendwocmd(cmd: str) -> None:
    """
    发送 WO 命令到游戏控制器

    参数:
        cmd: 要发送的 WO 命令
    """
    check_avaliable(game_control)
    game_control.sendwocmd(cmd)


def sendPacket(pktID: int, pkt: str) -> None:
    """
    发送数据包给游戏控制器

    参数:
        pktID: 数据包 ID
        pkt: 数据包内容
    """
    check_avaliable(game_control)
    game_control.sendPacket(pktID, pkt)


def rawText(playername: str, text: str) -> None:
    """
    向指定玩家发送原始文本消息

    参数:
        playername: 玩家名称
        text: 要发送的文本
    """
    game_control.say_to(playername, text)


def tellrawText(playername: str, title: str | None = None, text: str = "") -> None:
    """
    向指定玩家发送 tellraw 消息

    参数:
        playername: 玩家名称
        title: 标题文本（可选）
        text: 消息文本
    """
    if title is None:
        game_control.say_to(playername, text)
    else:
        game_control.say_to(playername, f"<{title}§r> {text}")


def get_all_player() -> list:
    """获取所有玩家列表"""
    check_avaliable(game_control)
    return game_control.allplayers


def is_op(playername: str) -> bool | None:
    """
    判断玩家是否为 OP

    参数:
        playername: 玩家名称
    """
    check_avaliable(game_control)
    if playername not in get_all_player():
        return False
    # 检测框架是否为"Frame"NeOmg
    if movent_frame.launcher.is_op is not None:
        return movent_frame.launcher.is_op(playername)
    return None


def find_key_from_value(dic: dict, val: Any) -> Optional[Any]:
    """
    从字典中根据值查找对应的键

    参数:
        dic: 目标字典
        val: 目标值
    """
    for k, v in dic.items():
        if v == val:
            return k
    return None


def get_robotname() -> str:
    """获取机器人名称。"""
    check_avaliable(game_control)
    return game_control.bot_name


def countdown(delay: int | float, msg: str | None = None) -> None:
    """
    倒计时函数

    参数:
        delay: 延迟时间，可以是整数或浮点数
        msg: 倒计时消息，可选参数，默认为"Countdown"
    """
    if msg is None:
        msg = "Countdown"
    delayStart = time.time()
    delayStop = delayStart + delay
    while delay >= 0:
        if delay >= 0:
            delay = delayStop - time.time()
        else:
            delay = 0
        Print.print_inf(f"{msg}: {delay:.2f}s", end="\r")
        time.sleep(0.01)
    print("", end="\n")


def getBlockTile(x: int, y: int, z: int) -> str:
    """
    获取指定坐标的方块的 ID

    参数:
        x: X 坐标
        y: Y 坐标
        z: Z 坐标
    """
    res = sendwscmd(f"/testforblock {x} {y} {z} air", True)
    if isinstance(res, type(None)):
        raise ValueError("Failed to get the block.")
    if res.SuccessCount:
        return "air"
    return res.OutputMessages[0].Parameters[4].strip("%tile.").strip(".name")


def getTickingAreaList() -> dict:
    """
    获取 tickingarea 列表

    异常:
        AttributeError: 获取 tickingarea 列表失败
    """
    result = {}
    cmd = sendwscmd("/tickingarea list all-dimensions", True)
    if cmd is not None:
        resultList = cmd.OutputMessages
    else:
        raise AttributeError("Failed to get the tickingarea list.")

    if resultList[0].Success is False:
        return result
    for tickareaData in resultList[1].Message.split("%dimension.dimensionName")[1:]:
        tickareaDimension = tickareaData.split(": \n")[0]
        tickareaList = tickareaData.split(": \n")[1].split("\n")
        for tickarea in tickareaList:
            if not tickarea:
                continue
            tickareaName = tickarea.split("- ", 1)[1].split(": ", 1)[0]
            tickareaPos = {
                "start": tickarea.split(" ")[2:5],
                "end": tickarea.split(" ")[6:9],
            }
            tickareaPos["start"].pop(1)
            tickareaPos["end"].pop(1)
            result[tickareaName] = {"dimension": tickareaDimension}
            result[tickareaName].update(tickareaPos)
    return result



"插件加载器框架"

import asyncio
import os
import traceback
from typing import TYPE_CHECKING, Any, Callable, Union, TypeVar

from ..color_print import Print
from .classic_plugin import (
    Plugin,
    add_plugin,
    add_plugin_as_api,
    _init_frame,
    _PLUGIN_CLS_TYPE,
)
from ..utils import Utils
from .injected_plugin import (
    execute_init,
    execute_player_prejoin,
    execute_player_join,
    execute_player_message,
    execute_death_message,
    execute_player_left,
    execute_frame_exit,
    execute_command_say,
    execute_repeat,
)
from ..plugin_load import (
    classic_plugin,
    injected_plugin,
    NON_FUNC,
    PluginAPINotFoundError,
    PluginAPIVersionError,
    auto_move_plugin_dir,
)
from ..constants import (
    PRG_NAME,
    TOOLDELTA_PLUGIN_DIR,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_INJECTED_PLUGIN,
)
from ..game_utils import _set_frame
from .injected_plugin.movent import set_frame as _set_frame_inj

if TYPE_CHECKING:
    from ..frame import ToolDelta

_TV = TypeVar("_TV")
_SUPER_CLS = TypeVar("_SUPER_CLS")


class PluginGroup:
    "插件组"

    plugins: list[Plugin] = []
    plugins_funcs: dict[str, list] = {
        "on_def": [],
        "on_inject": [],
        "on_player_prejoin": [],
        "on_player_join": [],
        "on_player_message": [],
        "on_player_death": [],
        "on_player_leave": [],
        "on_command": [],
        "on_frame_exit": [],
    }
    plugin_added_cache = {"packets": []}
    Agree_bot_patrol: list[bool] = []
    broadcast_evts_cache = {}

    def __init__(self):
        "初始化"
        self.listen_packet_ids = set()
        self._packet_funcs: dict[str, list[Callable]] = {}
        self._update_player_attributes_funcs: list[Callable] = []
        self._broadcast_listeners: dict[str, list[Callable]] = {}
        self.plugins_api: dict[str, Plugin] = {}
        self.normal_plugin_loaded_num = 0
        self.injected_plugin_loaded_num = 0
        self.loaded_plugins_name = []
        self.linked_frame: Union["ToolDelta", None] = None

    add_plugin = staticmethod(add_plugin)

    add_plugin_as_api = staticmethod(add_plugin_as_api)

    def add_packet_listener(self, pktID: int | list[int]):
        """
        添加数据包监听器
        将下面的方法作为一个 MC 数据包接收器
        Tips: 只能在插件主类里的函数使用此装饰器!

        Args:
            pktID (int | list[int]): 数据包 ID 或多个 ID

        Returns:
            Callable[[Callable], Callable]: 添加数据包监听器
        """

        def deco(func: Callable[[_SUPER_CLS, dict], bool]):
            if isinstance(pktID, int):
                self.plugin_added_cache["packets"].append((pktID, func))
            else:
                for i in pktID:
                    self.plugin_added_cache["packets"].append((i, func))
            return func

        return deco

    def add_broadcast_listener(self, evt_name: str):
        """
        添加广播事件监听器
        将下面的方法作为一个广播事件接收器
        Tips: 只能在插件主类里的函数使用此装饰器!

        Args:
            evt_name (str): 事件名

        Returns:
            Callable[[Callable], Callable]: 添加广播事件监听器

        原理:
        方法 1 广播：hi, what's ur name? 附加参数=english_only
            - 方法 2 接收到广播并被执行：方法 2(english_only) -> my name is Super. -> 收集表

        事件 1 获取到 收集表 作为返回：["my name is Super."]
        """

        def deco(
            func: Callable[[_SUPER_CLS, _TV], bool],
        ) -> Callable[[_SUPER_CLS, _TV], bool]:
            if self.broadcast_evts_cache.get(evt_name):
                self.broadcast_evts_cache[evt_name].append(func)
            else:
                self.broadcast_evts_cache[evt_name] = [func]
            return func

        return deco

    def broadcastEvt(self, evt_name: str, data: Any = None) -> list[Any]:
        """
        向全局广播一个特定事件，可以传入附加信息参数
        Args:
            evt_name (str): 事件名
            data (Any, optional): 附加信息参数
        Returns:
             list[Any]: 收集到的数据的列表 (如果接收到广播的方法返回了数据的话)
        """
        callback_list = []
        res = self._broadcast_listeners.get(evt_name)
        if res:
            for f in res:
                res2 = f(data)
                if res2:
                    callback_list.append(res2)
        return callback_list

    @staticmethod
    def help(plugin: Plugin) -> None:
        """
        查看插件帮助.
        常用于查看 get_plugin_api() 方法获取到的插件实例的帮助.
        """
        plugin_docs = "<plugins.help>: " + plugin.name + "开放的 API 接口说明:\n"
        for attr_name, attr in plugin.__dict__.items():
            if not attr_name.startswith("__") and attr.__doc__ is not None:
                plugin_docs += (
                    "\n §a"
                    + attr_name
                    + ":§f\n    "
                    + attr.__doc__.replace("\n", "\n    ")
                )
        Print.clean_print(plugin_docs)

    def checkSystemVersion(self, need_vers: tuple[int, int, int]):
        """检查 ToolDelta 系统的版本

        Args:
            need_vers (tuple[int, int, int]): 需要的版本

        Raises:
            self.linked_frame.SystemVersionException: 该组件需要的 ToolDelta 系统版本
        """
        if (
            self.linked_frame is not None
            and need_vers > self.linked_frame.sys_data.system_version
        ):
            raise self.linked_frame.SystemVersionException(
                f"该组件需要{PRG_NAME}为最低 {'.'.join([str(i) for i in self.linked_frame.sys_data.system_version])} 版本，请及时更新"
            )
        if self.linked_frame is None:
            raise ValueError(
                "无法检查 ToolDelta 系统版本，请确保已经加载了 ToolDelta 系统组件"
            )

    def get_plugin_api(
        self, apiName: str, min_version: tuple | None = None, force=True
    ) -> Any:
        """获取插件 API

        Args:
            apiName (str): 插件 API 名
            min_version (tuple | None, optional): API 最低版本 (若不填则默认不检查最低版本)
            force: 若为 False, 则在找不到插件 API 时不报错而是返回 None

        Raises:
            PluginAPIVersionError: 插件 API 版本错误
            PluginAPINotFoundError: 无法找到 API 插件

        Returns:
            Plugin: 插件 API
        """
        api = self.plugins_api.get(apiName)
        if api:
            if min_version and api.version < min_version:
                raise PluginAPIVersionError(apiName, min_version, api.version)
            return api
        if force:
            raise PluginAPINotFoundError(f"无法找到 API 插件：{apiName}")
        return None

    def set_frame(self, frame: "ToolDelta") -> None:
        """设置关联的系统框架"""
        self.linked_frame = frame
        _set_frame(frame)
        _set_frame_inj(frame)
        _init_frame(frame)

    def read_all_plugins(self) -> None:
        """读取所有插件

        Raises:
            SystemExit: 读取插件出现问题
        """
        if self.linked_frame is None or self.linked_frame.on_plugin_err is None:
            raise ValueError("无法读取插件，请确保系统已初始化")
        for fdir in os.listdir(TOOLDELTA_PLUGIN_DIR):
            if fdir not in (TOOLDELTA_CLASSIC_PLUGIN, TOOLDELTA_INJECTED_PLUGIN):
                auto_move_plugin_dir(fdir)
        try:
            classic_plugin.read_plugins(self)
            self.execute_def(self.linked_frame.on_plugin_err)
            asyncio.run(injected_plugin.load_plugin(self))
        except Exception as err:
            err_str = "\n".join(traceback.format_exc().split("\n")[1:])
            Print.print_err(f"加载插件出现问题：\n{err_str}")
            raise SystemExit from err

    def load_plugin_hot(self, plugin_name: str, plugin_type: str) -> None:
        """热加载插件

        Args:
            plugin_name (str): 插件名
            plugin_type (str): 插件类型
        """
        plugin = None
        if plugin_type == "classic":
            plugin = classic_plugin.load_plugin(self, plugin_name)
        elif plugin_type == "injected":
            asyncio.run(injected_plugin.load_plugin_file(plugin_name))
        # 检查是否有 on_def 成员再执行
        if plugin and hasattr(plugin, "on_def"):
            plugin.on_def()  # type: ignore
        Print.print_suc(f"成功热加载插件：{plugin_name}")

    def add_listen_packet_id(self, packetType: int) -> None:
        """添加数据包监听，仅在系统内部使用

        Args:
            packetType (int): 数据包 ID

        Raises:
            ValueError: 无法添加数据包监听，请确保已经加载了系统组件
        """
        if self.linked_frame is None:
            raise ValueError("无法添加数据包监听，请确保已经加载了系统组件")
        self.listen_packet_ids.add(packetType)
        self.linked_frame.link_game_ctrl.add_listen_pkt(packetType)

    def instant_plugin_api(self, api_cls: type[_PLUGIN_CLS_TYPE]) -> _PLUGIN_CLS_TYPE:
        """
        对外源导入 (import) 的 API 插件类进行类型实例化。
        可以使得你所使用的 IDE 对导入的插件 API 类进行识别和高亮其所含方法。

        Args:
            api_cls (type[_PLUGIN_CLS_TYPE]): 导入的 API 插件类

        Raises:
            ValueError: API 插件类未被注册

        Returns:
            _PLUGIN_CLS_TYPE: API 插件实例

        使用方法如下:
        ```python
            p_api = plugins.get_plugin_api("...")
            from outer_api import api_cls_xx
            p_api = plugins.instant_plugin_api(api_cls_xx)
        ```
        """
        for v in self.plugins_api.values():
            if isinstance(v, api_cls):
                return v
        raise ValueError(f"无法找到 API 插件类 {api_cls.__name__}, 有可能是还没有注册")

    def add_listen_packet_func(self, packetType: int, func: Callable) -> None:
        """添加数据包监听器，仅在系统内部使用

        Args:
            packetType (int): 数据包 ID
            func (Callable): 数据包监听器
        """
        if self._packet_funcs.get(str(packetType)):
            self._packet_funcs[str(packetType)].append(func)
        else:
            self._packet_funcs[str(packetType)] = [func]

    def add_broadcast_evt(self, evt: str, func: Callable) -> None:
        """添加广播事件监听器，仅在系统内部使用

        Args:
            evt (str): 事件名
            func (Callable): 事件监听器
        """
        if self._broadcast_listeners.get(evt):
            self._broadcast_listeners[evt].append(func)
        else:
            self._broadcast_listeners[evt] = [func]

    def _add_listen_update_player_attributes_func(self, func: Callable) -> None:
        """添加玩家属性更新监听器，仅在系统内部使用

        Args:
            func (Callable): 数据包监听器
        """
        self._update_player_attributes_funcs.append(func)

    def execute_def(
        self, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ) -> None:
        """执行插件的二次初始化方法

        Args:
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法。Defaults to NON_FUNC.

        Raises:
            SystemExit: 缺少前置
            SystemExit: 前置版本过低
        """
        try:
            for name, func in self.plugins_funcs["on_def"]:
                func()
        except PluginAPINotFoundError as err:
            name = err.name
            Print.print_err(f"插件 {name} 需要包含该种接口的前置组件：{err.name}")
            raise SystemExit from err
        except PluginAPIVersionError as err:
            Print.print_err(
                f"插件 {name} 需要该前置组件 {err.name} 版本：{err.m_ver}, 但是现有版本过低：{err.n_ver}"
            )
            raise SystemExit from err
        except Exception as err:
            onerr(name, err, traceback.format_exc())
            raise SystemExit

    def execute_init(
        self, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ) -> None:
        """执行插件的连接游戏后初始化方法

        Args:
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_inject"]:
            try:
                func()
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_init())
        Utils.createThread(asyncio.run, (execute_repeat(),))

    def execute_player_prejoin(
        self, player, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ) -> None:
        """执行玩家加入前的方法

        Args:
            player (_type_): 玩家
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_player_prejoin"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_player_prejoin(player))

    def execute_player_join(
        self, player: str, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ) -> None:
        """执行玩家加入的方法

        Args:
            player (str): 玩家
            onerr (Callable[[str, Exception, str], None], optional): q 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_player_join"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_player_join(player))

    def execute_player_message(
        self,
        player: str,
        msg: str,
        onerr: Callable[[str, Exception, str], None] = NON_FUNC,
    ) -> None:
        """执行玩家消息的方法

        Args:
            player (str): 玩家
            msg (str): 消息
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        pat = f"[{player}] "
        if msg.startswith(pat):
            msg = msg.strip(pat)
        for name, func in self.plugins_funcs["on_player_message"]:
            try:
                func(player, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_player_message(player, msg))

    def execute_player_leave(
        self, player: str, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ) -> None:
        """执行玩家离开的方法

        Args:
            player (str): 玩家
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_player_leave"]:
            try:
                func(player)
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_player_left(player))

    def execute_player_death(
        self,
        player: str,
        killer: str | None,
        msg: str,
        onerr: Callable[[str, Exception, str], None] = NON_FUNC,
    ):
        """执行玩家死亡的方法

        Args:
            player (str): 玩家
            killer (str | None): 击杀者
            msg (str): 消息
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_player_death"]:
            try:
                func(player, killer, msg)
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_death_message(player, killer, msg))

    def execute_command(
        self,
        name: str,
        msg: str,
        onerr: Callable[[str, Exception, str], None] = NON_FUNC,
    ) -> None:
        """执行命令 say 的方法

        Args:
            player (str): 玩家
            cmd (str): 命令
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for plugin_name, func in self.plugins_funcs["on_command"]:
            try:
                func(plugin_name, msg)
            except Exception as err:
                onerr(plugin_name, err, traceback.format_exc())
        asyncio.run(execute_command_say(name, msg))

    def execute_frame_exit(
        self, onerr: Callable[[str, Exception, str], None] = NON_FUNC
    ):
        """执行框架退出的方法

        Args:
            onerr (Callable[[str, Exception, str], None], optional): 插件出错时的处理方法
        """
        for name, func in self.plugins_funcs["on_frame_exit"]:
            try:
                func()
            except Exception as err:
                onerr(name, err, traceback.format_exc())
        asyncio.run(execute_frame_exit())

    def processPacketFunc(self, pktID: int, pkt: dict) -> bool:
        """处理数据包监听器

        Args:
            pktID (int): 数据包 ID
            pkt (dict): 数据包

        Returns:
            bool: 是否处理成功
        """
        d = self._packet_funcs.get(str(pktID))
        if d:
            for func in d:
                try:
                    res = func(pkt)
                    if res:
                        return True
                except Exception:
                    Print.print_err(f"插件方法 {func.__name__} 出错：")
                    Print.print_err(traceback.format_exc())
        return False


plugin_group = PluginGroup()



"插件市场客户端"

import os
import platform
import shlex
import shutil
import tempfile
import time
import traceback

import requests
import ujson as json

from . import urlmethod
from .cfg import Cfg
from .color_print import Print
from .constants import (
    PLUGIN_MARKET_SOURCE_OFFICIAL,
    TOOLDELTA_CLASSIC_PLUGIN,
    TOOLDELTA_INJECTED_PLUGIN,
)
from .plugin_load import PluginRegData
from .plugin_load.PluginGroup import plugin_group
from .utils import Utils

if platform.system().lower() == "windows":
    CLS_CMD = "cls"
else:
    CLS_CMD = "clear"


def clear_screen() -> None:
    "清屏"
    os.system(shlex.quote(CLS_CMD))


def path_dir(path: str) -> str | None:
    """获取路径的上一级目录

    Args:
        path (str): 路径

    Returns:
        str|None: 上一级目录
    """
    if "/" not in path:
        return None
    return "/".join(path.split("/")[:-1])


def url_join(*urls) -> str:
    """连接 URL

    Returns:
        str: 连接后的 URL
    """
    return "/".join(urls)


def get_json_from_url(url: str) -> dict:
    """从 URL 获取 JSON 数据

    Args:
        url (str): URL

    Raises:
        requests.RequestException: Url 请求失败
        requests.RequestException: 服务器返回了不正确的答复

    Returns:
        dict: JSON 数据
    """
    try:
        resp = requests.get(url, timeout=5)
        resp.raise_for_status()
        return resp.json()
    except requests.RequestException as exc:
        raise requests.RequestException("URL 请求失败") from exc
    except json.JSONDecodeError as exc:
        raise requests.RequestException(
            f"服务器返回了不正确的答复：{resp.text}"
        ) from exc


class PluginMarket:
    "插件市场类"

    def __init__(self):
        self.plugin_id_name_map: dict | None = None
        try:
            self.plugins_download_url = Cfg().get_cfg(
                "ToolDelta基本配置.json", {"插件市场源": str}
            )["插件市场源"]
        except Exception:
            self.plugins_download_url = PLUGIN_MARKET_SOURCE_OFFICIAL

    def enter_plugin_market(self, source_url: str | None = None, in_game=False) -> None:
        """进入插件市场

        Args:
            source_url (str | None, optional): 插件市场源
            in_game (bool, optional): 是否在游戏内
        """
        Print.clean_print("§6正在连接到插件市场..")
        self.plugin_id_name_map = self.get_plugin_id_name_map()
        CTXS = 12
        try:
            if isinstance(source_url, str):
                market_datas = self.get_datas_from_market(source_url)
            market_datas = self.get_datas_from_market()
            plugin_ids_map = self.plugin_id_name_map
            plugins_list = list(market_datas["MarketPlugins"].items())
            all_indexes = len(plugins_list)
            now_index = 0
            sum_pages = int((all_indexes - 1) / CTXS) + 1
            now_page = 0
            last_operation = ""
            while True:
                clear_screen()
                Print.print_inf(
                    market_datas["SourceName"] + ": " + market_datas["Greetings"],
                    need_log=False,
                )
                now_page = int(now_index / CTXS) + 1
                for i in range(now_index, now_index + CTXS):
                    if i in range(all_indexes):
                        plugin_id = plugins_list[i][0]
                        plugin_name = plugin_ids_map[plugin_id]
                        plugin_basic_datas = plugins_list[i][1]
                        if plugin_basic_datas["plugin-type"] == "classic":
                            plugin_type = "类式"
                        elif plugin_basic_datas["plugin-type"] == "injected":
                            plugin_type = "注入式"
                        else:
                            plugin_type = plugin_basic_datas["plugin-type"]
                        Print.print_inf(
                            (f" {i + 1}. §e{plugin_name} §av{plugin_basic_datas['version']} "
                             f"§b@{plugin_basic_datas['author']} §d{plugin_type}插件"),
                            need_log=False,
                        )
                    else:
                        Print.print_inf("")
                Print.print_inf(
                    f"§f第 {now_page} / {sum_pages} 页，输入 §b+§f/§b- §f翻页",
                    need_log=False,
                )
                Print.print_inf("§f输入插件序号选中插件并查看其下载页", need_log=False)
                last_operation = (
                    (
                        input(
                            Print.fmt_info(
                                "§f回车键继续上次操作，§bq§f 退出，请输入：", "§f 输入 "
                            )
                        )
                        or last_operation
                    )
                    .lower()
                    .strip()
                )
                if last_operation == "+":
                    now_index += CTXS
                elif last_operation == "-":
                    now_index -= CTXS
                elif last_operation == "q":
                    break
                else:
                    res = Utils.try_int(last_operation)
                    if res:
                        if res in range(1, all_indexes + 1):
                            plugin_data = self.get_plugin_data_from_market(
                                plugins_list[res - 1][0]
                            )
                            ok, pres = self.choice_plugin(plugin_data)
                            if ok:
                                if in_game:
                                    if (
                                        plugin_data.name
                                        not in plugin_group.loaded_plugins_name
                                    ):
                                        resp = (
                                            input(
                                                Print.fmt_info(
                                                    f"§f可以直接热加载该插件: {plugin_data.name}, 是否加载(§aY§f/§cN§f): "
                                                )
                                            )
                                            .strip()
                                            .lower()
                                        )
                                        if resp == "y":
                                            for i in pres:
                                                if (
                                                    i
                                                    not in plugin_group.loaded_plugins_name
                                                ):
                                                    try:
                                                        plugin_group.load_plugin_hot(
                                                            i.name, i.plugin_type
                                                        )
                                                    except BaseException as err:
                                                        Print.print_err(
                                                            f"插件热加载出现问题：{err}"
                                                        )
                                    else:
                                        Print.print_inf(
                                            "插件已存在，若要更新版本，请重启 ToolDelta",
                                            need_log=False,
                                        )
                                        r = input(
                                            Print.fmt_info(
                                                "§f输入 §cq §f退出, 其他则返回插件市场"
                                            )
                                        )
                                else:
                                    Print.print_inf(
                                        "下载插件后重启 ToolDelta 才能生效",
                                        need_log=False,
                                    )
                                r = input(
                                    Print.fmt_info(
                                        "§f输入 §cq §f退出, 其他则返回插件市场"
                                    )
                                )
                                if r.lower() == "q":
                                    break
                            else:
                                Print.print_inf("已取消。", need_log=False)
                                time.sleep(1)
                        else:
                            Print.print_err("超出序号范围")
                if now_index >= all_indexes:
                    now_index = 0
                elif now_index < 0:
                    now_index = max(now_index - CTXS, 0)
        except KeyError as err:
            Print.print_err(f"获取插件市场插件出现问题：键值对错误：{err}")
            return
        except requests.RequestException as err:
            Print.print_err(f"获取插件市场插件出现问题：{err}")
            return
        except Exception:
            Print.print_err("获取插件市场插件出现问题，报错如下：")
            Print.print_err(traceback.format_exc())
            return
        clear_screen()
        Print.clean_print("§a已从插件市场返回 ToolDelta 控制台。")

    def get_datas_from_market(self, source_url: str | None = None) -> dict:
        """
        从插件市场的 market_tree.json 获取数据

        Args:
            source_url (str | None, optional): 插件市场源，默认为 None, 有则替换整体插件市场源

        Returns:
            dict: 插件市场数据
        """
        if isinstance(source_url, str):
            self.plugins_download_url = self.plugins_download_url
        market_datas = get_json_from_url(
            url_join(self.plugins_download_url, "market_tree.json")
        )
        return market_datas

    def get_plugin_data_from_market(self, plugin_id: str) -> PluginRegData:
        """从插件市场获取单个插件数据

        Args:
            plugin_id (str): 插件 ID

        Raises:
            KeyError: 无法通过 ID 查找插件

        Returns:
            PluginRegData: 插件注册数据
        """
        if self.plugin_id_name_map is None:
            self.plugin_id_name_map = self.get_plugin_id_name_map()
        plugin_name = self.plugin_id_name_map.get(plugin_id)
        if plugin_name is None:
            raise KeyError(f"无法通过 ID: {plugin_id} 查找插件")
        data_url = self.plugins_download_url + "/" + plugin_name + "/datas.json"
        datas = get_json_from_url(data_url)
        return PluginRegData(plugin_name, datas)

    def choice_plugin(
        self, plugin_data: PluginRegData
    ) -> tuple[bool, list[PluginRegData]]:
        """选中插件进行介绍与操作

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Returns:
            tuple[bool, list[PluginRegData]]: 是否下载，下载的插件列表
        """
        pre_plugins_str = (
            ", ".join([f"{k}§7v{v}" for k, v in plugin_data.pre_plugins.items()])
            or "无"
        )
        clear_screen()
        Print.print_inf(
            f"{plugin_data.name} v{plugin_data.version_str}", need_log=False
        )
        Print.print_inf(
            f"作者：§f{plugin_data.author}§7, 版本：§f{plugin_data.version_str} §b{plugin_data.plugin_type_str}",
            need_log=False,
        )
        Print.print_inf(f"前置插件：§f{pre_plugins_str}", need_log=False)
        Print.print_inf(f"介绍：{plugin_data.description}", need_log=False)
        Print.print_inf("", need_log=False)
        res = (
            input(Print.fmt_info("§f下载 = §aY§f, 取消 = §cN§f, 请输入："))
            .lower()
            .strip()
        )
        if res == "y":
            Print.clean_print(f"§6正在下载插件：§f{plugin_data.name}", end="\r")
            pres = self.download_plugin(plugin_data)
            pres.reverse()
            return True, pres
        return False, []

    def get_plugin_id_name_map(self) -> dict:
        """获取插件 ID 与插件名的映射

        Returns:
            dict: 插件 ID 与插件名的映射
        """
        res = requests.get(
            self.plugins_download_url + "/plugin_ids_map.json", timeout=5
        )
        res.raise_for_status()
        res1: dict = json.loads(res.text)
        self.plugin_id_name_map = res1
        return res1

    def download_plugin(
        self, plugin_data: PluginRegData, with_pres=True, is_enabled=True
    ) -> list[PluginRegData]:
        """
        下载插件
        注意：只能传入由 `get_plugin_data_from_market()` 生成的数据

        Args:
            plugin_data (PluginRegData): 插件注册数据
            with_pres (bool): 是否一同下载前置插件
            is_enabled (bool): 下载的插件是否要自动禁用

        Raises:
            ValueError: 未知插件类型

        Returns:
            list[PluginRegData]: 本插件和其前置插件的注册数据列表
        """
        if self.plugin_id_name_map is None:
            self.plugin_id_name_map = self.get_plugin_id_name_map()
        pres = [plugin_data]
        download_paths = self.find_dirs(plugin_data)
        if with_pres:
            for plugin_id in plugin_data.pre_plugins:
                plugin_name = self.plugin_id_name_map[plugin_id]
                Print.clean_print(
                    f"§6正在下载 §f{plugin_data.name}§6 的前置插件 §f{plugin_name}"
                )
                plugin_datas = self.get_plugin_data_from_market(plugin_id)
                pres += self.download_plugin(plugin_datas)
        cache_dir = tempfile.mkdtemp()
        Print.clean_print(
            f"§6正在下载插件 §f{plugin_data.name}§6.." + " " * 15, end="\r"
        )
        match plugin_data.plugin_type:
            case "classic":
                download_path = os.path.join("插件文件", TOOLDELTA_CLASSIC_PLUGIN)
            case "injected":
                download_path = os.path.join("插件文件", TOOLDELTA_INJECTED_PLUGIN)
            case _:
                raise ValueError(
                    f"未知插件类型：{plugin_data.plugin_type}, 你可能需要通知 ToolDelta 项目开发组解决"
                )
        try:
            os.makedirs(os.path.join(cache_dir, plugin_data.name), exist_ok=True)
            all_files_len = len(download_paths)
            for i, paths in enumerate(download_paths):
                Print.clean_print(
                    f"正在下载附带文件 ({i} / {all_files_len}) 请稍后...", end="\r"
                )
                if not paths.strip():
                    # 该路径为空
                    continue
                url = url_join(self.plugins_download_url, paths)
                # 按照插件类型选择下载到的文件夹
                path_last = path_dir(paths)
                if path_last is not None:
                    # 自动创建文件夹
                    folder_path = os.path.join(cache_dir, path_last)
                    os.makedirs(folder_path, exist_ok=True)
                urlmethod.download_unknown_file(url, os.path.join(cache_dir, paths))
            # 将缓存文件夹的插件移动到本文件夹
            target_path = download_path
            os.makedirs(target_path, exist_ok=True)
            # 制作所需的目录结构
            for root, _, files in os.walk(cache_dir):
                for filename in files:
                    source_file = os.path.join(root, filename)
                    target_file = os.path.join(
                        target_path, os.path.relpath(source_file, cache_dir)
                    )
                    os.makedirs(os.path.dirname(target_file), exist_ok=True)
                    shutil.move(source_file, target_file)
            if not is_enabled:
                shutil.rmtree(os.path.join(target_path, plugin_data.name + "+disabled"))
                os.rename(
                    os.path.join(target_path, plugin_data.name),
                    os.path.join(target_path, plugin_data.name + "+disabled"),
                )
            from .plugin_manager import plugin_manager

            plugin_data = self.get_plugin_data_from_market(plugin_data.plugin_id)
            plugin_data.is_enabled = is_enabled
            plugin_manager.push_plugin_reg_data(plugin_data)
            Print.clean_print(
                f"§a成功下载插件 §f{plugin_data.name}§a 至插件文件夹" + " " * 15
            )
        finally:
            shutil.rmtree(cache_dir)
        return pres

    def find_dirs(self, plugin_data: PluginRegData) -> list[str]:
        """查找插件目录

        Args:
            plugin_data (PluginRegData): 插件注册数据

        Raises:
            KeyError: 目录结构错误
            eExceptionrr: 获取插件市场插件目录结构出现问题

        Returns:
            list[str]: 插件目录列表
        """
        try:
            data = get_json_from_url(
                url_join(self.plugins_download_url, "directory.json")
            )
            data_list = []
            for folder, files in data.items():
                if plugin_data.name == folder.split("/")[0]:
                    # 展开
                    for file in files:
                        data_list.append(folder + r"/" + file)
            return data_list
        except KeyError as err:
            Print.print_err(
                f"获取插件市场插件目录结构出现问题：无法找到 {err}, 有可能是未来得及更新目录"
            )
            raise KeyError(f"无法找到 {err}, 有可能是未来得及更新目录") from err
        except Exception as err:
            Print.print_err(f"获取插件市场插件目录结构出现问题：{err}")
            raise KeyError(f"获取插件市场插件目录结构出现问题：{err}") from err

    def get_latest_plugin_version(self, plugin_id: str) -> str:
        """获取最新插件版本

        Args:
            plugin_id (str): 插件 ID

        Raises:
            KeyError: 无法通过 ID 获取最新插件版本

        Returns:
            str: 最新插件版本
        """
        result = get_json_from_url(
            url_join(self.plugins_download_url, "latest_versions.json")
        ).get(plugin_id)
        if isinstance(result, str):
            return result
        raise KeyError(f"无法通过 ID: {plugin_id} 获取最新插件版本")


market = PluginMarket()



r"""
# ToolDelta: NEMC Rental Server R-Plugin Loader

网易我的世界手机版 租赁服机器人式插件加载器

- 进入 ToolDelta 界面:

   ```python
   from tooldelta import client_title

   client_title()
   ```
- 快速启动 ToolDelta:

   ```python
   from tooldelta import start_tool_delta

   start_tool_delta()
   ```
- 快速创建插件所需库环境

   ```python
   from tooldelta import *
   ```
"""

from .color_print import Print  # noqa: F401
from .frame import Config, GameCtrl, ToolDelta, Utils  # noqa: F401
from .launch_options import client_title  # noqa: F401
from typing import TYPE_CHECKING  # noqa: F401
from .plugin_load.PluginGroup import Plugin  # noqa: F401
from .starter import plugin_group as plugins  # noqa: F401
from .starter import safe_jump, start_tool_delta, tooldelta  # noqa: F401

# 重定向
Builtins = Utils
Frame = ToolDelta



"ToolDelta 注入式插件"

import asyncio
from dataclasses import dataclass
import os
import sys
import importlib

from typing import TYPE_CHECKING, Callable, List, Tuple
from ...color_print import Print
from ...plugin_load import (
    plugin_is_enabled,
    PluginAPINotFoundError,
    PluginAPIVersionError,
)

if TYPE_CHECKING:
    from tooldelta.plugin_load.PluginGroup import PluginGroup

# 定义插件处理函数列表
player_message_funcs: dict[Callable, int | None] = {}
player_prejoin_funcs: dict[Callable, int | None] = {}
player_join_funcs: dict[Callable, int | None] = {}
player_left_funcs: dict[Callable, int | None] = {}
player_death_funcs: dict[Callable, int | None] = {}
commmand_message_funcs: dict[Callable, int | None] = {}
repeat_funcs: dict[Callable, int | float] = {}
init_plugin_funcs: dict[Callable, int | None] = {}
frame_exit_funcs: dict[Callable, int | None] = {}


def player_message(priority: int | None = None) -> Callable:
    """载入处理玩家消息

    Args:
        priority (int, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        player_message_funcs[func] = priority
        return func

    return decorator


def player_prejoin(priority: int | None = None) -> Callable[["player_name"], None]:
    """载入处理玩家加入前事件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        player_prejoin_funcs[func] = priority
        return func

    return decorator


def player_join(priority: int | None = None) -> Callable:
    """载入处理玩家加入事件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        player_join_funcs[func] = priority
        return func

    return decorator


def player_left(priority: int | None = None) -> Callable:
    """载入处理玩家离开事件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        player_left_funcs[func] = priority
        return func

    return decorator


def player_death(priority: int | None = None) -> Callable:
    """载入处理玩家死亡事件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        player_death_funcs[func] = priority
        return func

    return decorator


def init(priority: int | None = None) -> Callable:
    """载入机器人进入游戏后初始化插件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        init_plugin_funcs[func] = priority
        return func

    return decorator


def frame_exit(priority: int | None = None) -> Callable:
    """载入处理框架退出事件的插件

    Args:
        priority (int | None, optional): 插件优先级

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        frame_exit_funcs[func] = priority
        return func

    return decorator


def repeat(retime: int | float = 5) -> Callable:
    """载入重复任务

    Args:
        retime (int, optional): 重复时间

    Returns:
        Callable: 插件处理函数
    """

    def decorator(func):
        repeat_funcs[func] = retime
        return func

    return decorator


async def command_say(priority: int | None = None) -> Callable:
    """载入处理命令消息

    Args:
        name (str): 命令名
        priority (int | None, optional): 插件优先级
    """

    def decorator(func):
        commmand_message_funcs[func] = priority
        return func

    return decorator


async def repeat_task(func: Callable, time: int | float) -> None:
    """执行重复任务（执行完等待一段时间再执行）

    Args:
        func (Callable): 定时执行的函数
        time (int | float): 重复时间
    """
    while True:
        await asyncio.sleep(time)
        # 防止出错
        try:
            await func()
        except Exception as e:
            Print.print_err(f"repeat_task error: {e}")


async def execute_asyncio_task(func_dict: dict, *args, **kwargs) -> None:
    """执行异步任务

    Args:
        func_dict (dict): 函数字典
    """
    tasks: List[Tuple[int, asyncio.Task]] = []
    none_tasks: List[Tuple[None, asyncio.Task]] = []
    # 将任务添加到 tasks 列表或 none_tasks 列表中
    for func, priority in func_dict.items():
        if priority is not None:
            tasks.append((priority, asyncio.create_task(func(*args, **kwargs))))
        else:
            none_tasks.append((priority, asyncio.create_task(func(*args, **kwargs))))

    # 按优先级对非 None 任务排序
    tasks.sort(key=lambda x: x[0])

    # 将 none_tasks 列表附加到已排序的任务列表的末尾
    task_list = none_tasks + tasks

    await asyncio.gather(*[task[1] for task in task_list])


async def execute_init() -> None:
    """执行初始化插件函数"""
    await execute_asyncio_task(init_plugin_funcs)


async def run_repeat():
    """执行重复任务"""
    # 为字典中的每一个函数创建一个循环特定时间的任务
    tasks = []
    for func, time in repeat_funcs.items():
        tasks.append(repeat_task(func, time))
    # 并发执行所有任务
    await asyncio.gather(*tasks)


async def safe_jump():
    """安全跳出重复任务"""
    try:
        main_task.cancel()
    except NameError:
        return


main_task: asyncio.Task


@dataclass(order=True)
class command_message_info:
    """命令消息信息"""

    name: str
    message: str


@dataclass(order=True)
class player_name:
    """玩家名字"""

    playername: str


@dataclass(order=True)
class player_message_info(player_name):
    """玩家消息信息"""

    message: str


@dataclass(order=True)
class player_death_info(player_name):
    """玩家死亡信息"""

    message: str
    killer: str | None = None


async def execute_repeat() -> None:
    """执行重复任务"""
    # skipcq: PYL-W0603
    global main_task
    main_task = asyncio.create_task(run_repeat())
    try:
        await main_task
    except asyncio.CancelledError:
        Print.print_suc("重复任务 repeat_task 已退出！")


async def execute_player_message(playername: str, message: str) -> None:
    """执行玩家消息处理函数

    Args:
        playername (str): 玩家名字
        message (str): 消息
    """
    await execute_asyncio_task(
        player_message_funcs,
        player_message_info(playername=playername, message=message),
    )


async def execute_death_message(
    playername: str, killer: str | None, message: str
) -> None:
    """执行玩家死亡处理函数

    Args:
        playername (str): 玩家名字
        killer (str | None): 凶手
        message (str): 消息
    """
    await execute_asyncio_task(
        player_death_funcs,
        player_death_info(playername=playername, killer=killer, message=message),
    )


async def execute_player_join(playername: str) -> None:
    """执行玩家加入处理函数

    Args:
        playername (str): 玩家名字
    """
    await execute_asyncio_task(player_join_funcs, player_name(playername=playername))


async def execute_player_prejoin(playername: str) -> None:
    """执行玩家加入前处理函数

    Args:
        playername (str): 玩家名字
    """
    await execute_asyncio_task(player_prejoin_funcs, player_name(playername=playername))


async def execute_player_left(playername: str) -> None:
    """执行玩家离开处理函数

    Args:
        playername (str): 玩家名字
    """
    await execute_asyncio_task(player_left_funcs, player_name(playername=playername))


async def execute_command_say(name: str, message: str) -> None:
    """执行命令消息处理函数

    Args:
        message (str): 消息 say
    """
    await execute_asyncio_task(
        commmand_message_funcs, command_message_info(name=name, message=message)
    )


async def execute_frame_exit() -> None:
    """执行框架退出处理函数"""
    await execute_asyncio_task(frame_exit_funcs)


class PluginMetadata:
    """插件元数据"""

    def __init__(
        self,
        name,
        author,
        description,
        version,
        usage,
        homepage,
    ):
        self.name = name
        self.author = author
        self.description = description
        self.usage = usage
        self.version = version
        self.homepage = homepage


def create_plugin_metadata(metadata_dict: dict) -> PluginMetadata:
    """创建插件元数据

    Args:
        metadata_dict (dict): 插件元数据字典

    Returns:
        PluginMetadata: 插件元数据
    """
    name = metadata_dict.get("name", "未命名插件")
    version = metadata_dict.get("version", "1.0")
    description = metadata_dict.get("description", "未知插件")
    usage = metadata_dict.get("usage", "")
    author = metadata_dict.get("author", "未知")
    homepage = metadata_dict.get("homepage", "")

    return PluginMetadata(name, author, description, version, usage, homepage)


async def load_plugin_file(file: str) -> PluginMetadata:
    """加载插件文件

    Args:
        file (str): 插件文件名

    Returns:
        PluginMetadata: 插件元数据
    """
    try:
        # 导入插件模块
        sys.path.append(os.path.join("插件文件", "ToolDelta注入式插件"))
        plugin_module = importlib.import_module(file)
        meta_data = create_plugin_metadata(
            getattr(plugin_module, "__plugin_meta__", {"name": file})
        )
        # 获取插件元数据
        return meta_data
    except PluginAPIVersionError as err:
        Print.print_err(
            f"插件 {file} 加载出现问题：需要 {err.name} 的 API 最低版本为 {err.m_ver}, 实际上只有 {err.n_ver}"
        )
        raise
    except PluginAPINotFoundError as err:
        Print.print_err(f"插件 {file} 加载出现问题：需要前置插件 API {err.name}")
        raise


async def load_plugin(plugin_grp: "PluginGroup") -> None:
    """加载插件

    Args:
        plugin_grp (PluginGroup): 插件组
    """
    tasks = []

    # 读取本目录下的文件夹名字
    PLUGIN_PATH = os.path.join(os.getcwd(), "插件文件", "ToolDelta注入式插件")
    for file in os.listdir(PLUGIN_PATH):
        if not plugin_is_enabled(file):
            continue
        if os.path.isdir(os.path.join(PLUGIN_PATH, file)):
            plugin_grp.injected_plugin_loaded_num += 1
            plugin_grp.loaded_plugins_name.append(file)
            task = asyncio.create_task(load_plugin_file(file))
            tasks.append(task)

    # 顺序加载插件并收集插件元数据
    all_plugin_metadata = []
    for task in tasks:
        plugin_metadata = await task
        all_plugin_metadata.append(plugin_metadata)

    # 打印所有插件的元数据
    for metadata in all_plugin_metadata:
        Print.print_suc(
            f"成功载入插件 {metadata.name} 版本：{metadata.version} 作者：{metadata.author}"
        )
